@@ -60,6 +60,7 @@ GenericMMC::GenericMMC(ScsiIf *scsiIf, unsigned long options)
   driverName_ = "Generic SCSI-3/MMC - Version 2.0";
   
   speed_ = 0;
+  rspeed_ = 0;
   simulate_ = 1;
   encodingMode_ = 1;
 
@@ -153,7 +154,7 @@ int GenericMMC::speed(int s)
 {
   speed_ = s;
 
-  if (selectSpeed(0) != 0)
+  if (selectSpeed() != 0)
     return 1;
   
   return 0;
@@ -166,7 +167,7 @@ int GenericMMC::speed()
   delete driveInfo_;
   driveInfo_ = NULL;
 
-  if ((di = driveInfo(1)) == NULL) {
+  if ((di = driveInfo(true)) == NULL) {
     return 0;
   }
 
@@ -174,6 +175,34 @@ int GenericMMC::speed()
 
 }
 
+// sets fspeed
+// return: true: OK
+//         false: illegal speed
+bool GenericMMC::rspeed(int s)
+{
+  rspeed_ = s;
+
+  if (selectSpeed() != 0)
+    return false;
+  
+  return true;
+}
+
+int GenericMMC::rspeed()
+{
+  const DriveInfo *di;
+
+  delete driveInfo_;
+  driveInfo_ = NULL;
+
+  if ((di = driveInfo(true)) == NULL) {
+    return 0;
+  }
+
+  return speed2Mult(di->currentReadSpeed);
+}
+
+
 // loads ('unload' == 0) or ejects ('unload' == 1) tray
 // return: 0: OK
 //         1: scsi command failed
@@ -310,7 +339,7 @@ int GenericMMC::blankDisk(BlankingMode mode)
 // sets read/write speed and simulation mode
 // return: 0: OK
 //         1: scsi command failed
-int GenericMMC::selectSpeed(int readSpeed)
+int GenericMMC::selectSpeed()
 {
   unsigned char cmd[12];
   int spd;
@@ -320,12 +349,12 @@ int GenericMMC::selectSpeed(int readSpeed)
   cmd[0] = 0xbb; // SET CD SPEED
 
   // select maximum read speed
-  if (readSpeed == 0) {
+  if (rspeed_ == 0) {
     cmd[2] = 0xff;
     cmd[3] = 0xff;
   }
   else {
-    spd = mult2Speed(readSpeed);
+    spd = mult2Speed(rspeed_);
     cmd[2] = spd >> 8;
     cmd[3] = spd;
   }
@@ -378,7 +407,15 @@ int GenericMMC::getSessionInfo()
 
   if (leadInStart_.lba() >= Msf(80, 0, 0).lba()) {
     leadInLen_ = 450000 - leadInStart_.lba();
-    leadOutLen_ = Msf(1, 30, 0).lba(); // 90 seconds lead-out
+
+    if (fullBurn_) {
+    	leadOutLen_ = (userCapacity_ ? Msf(userCapacity_, 0, 0).lba() : diskInfo_.capacity) + Msf(1, 30, 0).lba() - toc_->length().lba() - diskInfo_.thisSessionLba - 150; // Fill all rest space <vladux>
+	if (leadOutLen_ < Msf(1, 30, 0).lba()) {
+		leadOutLen_ = Msf(1, 30, 0).lba(); // 90 seconds lead-out
+	}
+    } else {
+    	leadOutLen_ = Msf(1, 30, 0).lba(); // 90 seconds lead-out
+    }
   }
   else {
     leadInLen_ = Msf(1, 0, 0).lba();
@@ -480,7 +517,7 @@ int GenericMMC::setWriteParameters(unsigned long variant)
   }
 
   const DriveInfo *di;
-  if ((di = driveInfo(1)) != NULL) {
+  if ((di = driveInfo(true)) != NULL) {
     if (di->burnProof) {
       // This drive has BURN-Proof function.
       // Enable it unless explicitly disabled.
@@ -651,7 +688,7 @@ int GenericMMC::getStartOfSession(long *lba)
 
 static unsigned char leadInOutDataMode(TrackData::Mode mode)
 {
-  unsigned char ret;
+  unsigned char ret = 0;
 
   switch (mode) {
   case TrackData::AUDIO:
@@ -747,7 +784,7 @@ unsigned char *GenericMMC::createCueSheet(unsigned long variant,
     len += t->nofIndices(); // entry for each index increment
   }
 
-  cueSheet = new (unsigned char)[len * 8];
+  cueSheet = new unsigned char[len * 8];
   n = 0;
 
   if (toc_->leadInMode() == TrackData::AUDIO) {
@@ -1062,7 +1099,7 @@ int GenericMMC::initDao(const Toc *toc)
     return 1;
   }
 
-  if (selectSpeed(0) != 0 ||
+  if (selectSpeed() != 0 ||
       getSessionInfo() != 0) {
     return 1;
   }
@@ -1602,7 +1639,7 @@ int GenericMMC::readIsrc(int trackNr, char *buf)
 
   cmd[0] = 0x42; // READ SUB-CHANNEL
   cmd[2] = 1 << 6;
-  cmd[3] = 0x03; // get media catalog number
+  cmd[3] = 0x03; // get ISRC
   cmd[6] = trackNr;
   cmd[8] = 24;   // transfer length
 
@@ -1629,7 +1666,7 @@ int GenericMMC::analyzeTrack(TrackData::Mode mode, int trackNr, long startLba,
   int ret;
   int noScan = 0;
 
-  selectSpeed(0);
+  selectSpeed();
 
   if ((readCapabilities_ & CDR_AUDIO_SCAN_CAP) == 0) {
     ret = analyzeTrackSearch(mode, trackNr, startLba, endLba, indexIncrements,
@@ -1663,7 +1700,7 @@ int GenericMMC::readSubChannels(TrackData::SubChannelMode sm,
   int retries = 5;
   unsigned char cmd[12];
   int i;
-  long blockLen;
+  long blockLen = 0;
   unsigned long subChanMode = 0;
 
   cmd[0] = 0xbe;  // READ CD
@@ -1872,7 +1909,7 @@ int GenericMMC::getFeature(unsigned int feature, unsigned char *buf,
   if (len > bufLen)
     len = bufLen;
 
-  data = new (unsigned char)[len + 8];
+  data = new unsigned char[len + 8];
 
   cmd[7] = (len + 8) >> 8;
   cmd[8] = (len + 8);
@@ -1906,7 +1943,7 @@ int GenericMMC::getFeature(unsigned int feature, unsigned char *buf,
   return 0;
 }
 
-const DriveInfo *GenericMMC::driveInfo(int showErrorMsg)
+const DriveInfo *GenericMMC::driveInfo(bool showErrorMsg)
 {
   unsigned char mp[32];
 
@@ -2014,7 +2051,7 @@ CdRawToc *GenericMMC::getRawToc(int sessionNr, int *len)
   
   message(4, "Raw toc data len: %d", dataLen);
 
-  data = new (unsigned char)[dataLen];
+  data = new unsigned char[dataLen];
   
   // read disk toc
   cmd[7] = dataLen >> 8;
@@ -2486,8 +2523,14 @@ int GenericMMC::readCdTest(long lba, long len, int subChanMode) const
 
 	  chan.type(SubChannel::QMODE1DATA);
 
-	  long pqlba =
-	    Msf(chan.amin(), chan.asec(), chan.aframe()).lba() - 150;
+	  int min = chan.amin();
+	  int sec = chan.asec();
+	  int frac = chan.aframe();
+
+	  if ((frac >= 0 && frac < 75) &&
+          (sec >= 0 && sec < 60) &&
+          (min >= 0)) {
+	  long pqlba = Msf(min, sec, frac).lba() - 150;
 
 	  long diff = pqlba - lba;
 	  if (diff < 0)
@@ -2496,6 +2539,7 @@ int GenericMMC::readCdTest(long lba, long len, int subChanMode) const
 	  if (diff < 20) {
 	    pqSubChanBcdOk++;
 	  }
+	  }
 	}
 
 	if (buf[7] < 100 && buf[8] < 60 && buf[9] < 75) {
@@ -2708,7 +2752,7 @@ int GenericMMC::RicohGetWriteOptions()
   if (mp[2] & (1 << 5))
     driveInfo_->ricohJustSpeed = 1;
 
-  if (mp[2] & (1 << 1))
+  if (mp[2] & 0x01)
     driveInfo_->ricohJustLink = 1;
 
   return 0;
