@@ -1,7 +1,7 @@
-/* @(#)scsi-bsd.c	1.25 99/09/07 Copyright 1997 J. Schilling */
+/* @(#)scsi-bsd.c	1.42 04/01/15 Copyright 1997 J. Schilling */
 #ifndef lint
 static	char __sccsid[] =
-	"@(#)scsi-bsd.c	1.25 99/09/07 Copyright 1997 J. Schilling";
+	"@(#)scsi-bsd.c	1.42 04/01/15 Copyright 1997 J. Schilling";
 #endif
 /*
  *	Interface for the NetBSD/FreeBSD/OpenBSD generic SCSI implementation.
@@ -12,6 +12,12 @@ static	char __sccsid[] =
  *	similar to the SCSI command transport of the
  *	6 years older scg driver.
  *
+ *	Warning: you may change this source, but if you do that
+ *	you need to change the _scg_version and _scg_auth* string below.
+ *	You may not return "schily" for an SCG_AUTHOR request anymore.
+ *	Choose your name instead of "schily" and make clear that the version
+ *	string is related to a modified source.
+ *
  *	Copyright (c) 1997 J. Schilling
  */
 /*
@@ -25,9 +31,9 @@ static	char __sccsid[] =
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; see the file COPYING.  If not, write to
- * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
+ * You should have received a copy of the GNU General Public License along with
+ * this program; see the file COPYING.  If not, write to the Free Software
+ * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
 #ifndef HAVE_CAMLIB_H
@@ -35,6 +41,15 @@ static	char __sccsid[] =
 #undef	sense
 #include <sys/scsiio.h>
 
+/*
+ *	Warning: you may change this source, but if you do that
+ *	you need to change the _scg_version and _scg_auth* string below.
+ *	You may not return "schily" for an SCG_AUTHOR request anymore.
+ *	Choose your name instead of "schily" and make clear that the version
+ *	string is related to a modified source.
+ */
+LOCAL	char	_scg_trans_version[] = "scsi-bsd.c-1.42";	/* The version for this transport*/
+
 #define	MAX_SCG		16	/* Max # of SCSI controllers */
 #define	MAX_TGT		16
 #define	MAX_LUN		8
@@ -42,7 +57,7 @@ static	char __sccsid[] =
 struct scg_local {
 	short	scgfiles[MAX_SCG][MAX_TGT][MAX_LUN];
 };
-#define scglocal(p)	((struct scg_local *)((p)->local)) 
+#define	scglocal(p)	((struct scg_local *)((p)->local))
 
 /*#define	MAX_DMA_BSD	(32*1024)*/
 #define	MAX_DMA_BSD	(60*1024)	/* More seems to make problems */
@@ -59,23 +74,66 @@ struct scg_local {
 #define	SADDR_TARGET(a)	(SADDR_ISSCSI(a)?(a).addr.scsi.target:(a).addr.atapi.atbus*2+(a).addr.atapi.drive)
 #define	SADDR_LUN(a)	(SADDR_ISSCSI(a)?(a).addr.scsi.lun:0)
 #else
+
+#if	defined(__OpenBSD__) && defined(TYPE_ATAPI)
+#define	SADDR_ISSCSI(a)	((a).type == TYPE_SCSI)
+#else
 #define	SADDR_ISSCSI(a)	(1)
+#endif
 
 #define	SADDR_BUS(a)	(a).scbus
 #define	SADDR_TARGET(a)	(a).target
 #define	SADDR_LUN(a)	(a).lun
 #endif	/* __NetBSD__ && TYPE_ATAPI */
 
-LOCAL	BOOL	scsi_setup	__PR((SCSI *scgp, int f, int busno, int tgt, int tlun));
+LOCAL	BOOL	scg_setup	__PR((SCSI *scgp, int f, int busno, int tgt, int tlun));
 
-EXPORT int
-scsi_open(scgp, device, busno, tgt, tlun)
+/*
+ * Return version information for the low level SCSI transport code.
+ * This has been introduced to make it easier to trace down problems
+ * in applications.
+ */
+LOCAL char *
+scgo_version(scgp, what)
+	SCSI	*scgp;
+	int	what;
+{
+	if (scgp != (SCSI *)0) {
+		switch (what) {
+
+		case SCG_VERSION:
+			return (_scg_trans_version);
+		/*
+		 * If you changed this source, you are not allowed to
+		 * return "schily" for the SCG_AUTHOR request.
+		 */
+		case SCG_AUTHOR:
+			return (_scg_auth_schily);
+		case SCG_SCCS_ID:
+			return (__sccsid);
+		}
+	}
+	return ((char *)0);
+}
+
+LOCAL int
+scgo_help(scgp, f)
+	SCSI	*scgp;
+	FILE	*f;
+{
+	__scg_help(f, "SCIOCCOMMAND", "SCSI for devices known by *BSD",
+		"", "device or bus,target,lun", "/dev/rcd0a:@ or 1,2,0", FALSE, TRUE);
+	return (0);
+}
+
+LOCAL int
+scgo_open(scgp, device)
 	SCSI	*scgp;
 	char	*device;
-	int	busno;
-	int	tgt;
-	int	tlun;
 {
+		int	busno	= scg_scsibus(scgp);
+		int	tgt	= scg_target(scgp);
+		int	tlun	= scg_lun(scgp);
 	register int	f;
 	register int	b;
 	register int	t;
@@ -93,13 +151,13 @@ scsi_open(scgp, device, busno, tgt, tlun)
 	}
 
 	if (scgp->local == NULL) {
-		scgp->local = malloc(sizeof(struct scg_local));
+		scgp->local = malloc(sizeof (struct scg_local));
 		if (scgp->local == NULL)
 			return (0);
 
-		for (b=0; b < MAX_SCG; b++) {
-			for (t=0; t < MAX_TGT; t++) {
-				for (l=0; l < MAX_LUN ; l++)
+		for (b = 0; b < MAX_SCG; b++) {
+			for (t = 0; t < MAX_TGT; t++) {
+				for (l = 0; l < MAX_LUN; l++)
 					scglocal(scgp)->scgfiles[b][t][l] = (short)-1;
 			}
 		}
@@ -110,18 +168,20 @@ scsi_open(scgp, device, busno, tgt, tlun)
 
 	if (busno >= 0 && tgt >= 0 && tlun >= 0) {
 
-		sprintf(devname, "/dev/su%d-%d-%d", busno, tgt, tlun);
+		js_snprintf(devname, sizeof (devname),
+				"/dev/su%d-%d-%d", busno, tgt, tlun);
 		f = open(devname, O_RDWR);
 		if (f < 0) {
 			goto openbydev;
 		}
 		scglocal(scgp)->scgfiles[busno][tgt][tlun] = f;
-		return(1);
+		return (1);
 
-	} else for (b=0; b < MAX_SCG; b++) {
-		for (t=0; t < MAX_TGT; t++) {
-			for (l=0; l < MAX_LUN ; l++) {
-				sprintf(devname, "/dev/su%d-%d-%d", b, t, l);
+	} else for (b = 0; b < MAX_SCG; b++) {
+		for (t = 0; t < MAX_TGT; t++) {
+			for (l = 0; l < MAX_LUN; l++) {
+				js_snprintf(devname, sizeof (devname),
+							"/dev/su%d-%d-%d", b, t, l);
 				f = open(devname, O_RDWR);
 /*				error("open (%s) = %d\n", devname, f);*/
 
@@ -136,7 +196,7 @@ scsi_open(scgp, device, busno, tgt, tlun)
 						return (0);
 					}
 				} else {
-					if (scsi_setup(scgp, f, b, t, l))
+					if (scg_setup(scgp, f, b, t, l))
 						nopen++;
 				}
 			}
@@ -153,31 +213,37 @@ openbydev:
 		if (device == NULL || device[0] == '\0')
 			return (0);
 		f = open(device, O_RDWR);
-		if (f < 0)
+		if (f < 0) {
+			if (scgp->errstr)
+				js_snprintf(scgp->errstr, SCSI_ERRSTR_SIZE,
+					"Cannot open '%s'",
+					device);
 			return (0);
+		}
 		if (tgt == -2) {
 			if (ioctl(f, SCIOCIDENTIFY, &saddr) < 0) {
 				close(f);
 				errno = EINVAL;
 				return (0);
 			}
-			scgp->scsibus = busno	= SADDR_BUS(saddr);
-			scgp->target  = tgt	= SADDR_TARGET(saddr);
+			busno	= SADDR_BUS(saddr);
+			tgt	= SADDR_TARGET(saddr);
 			if ((tlun >= 0) && (tlun != SADDR_LUN(saddr))) {
 				close(f);
 				errno = EINVAL;
 				return (0);
 			}
-			scgp->lun     = tlun	= SADDR_LUN(saddr);
+			tlun	= SADDR_LUN(saddr);
+			scg_settarget(scgp, busno, tgt, tlun);
 		}
-		if (scsi_setup(scgp, f, busno, tgt, tlun))
+		if (scg_setup(scgp, f, busno, tgt, tlun))
 			nopen++;
 	}
 	return (nopen);
 }
 
-EXPORT int
-scsi_close(scgp)
+LOCAL int
+scgo_close(scgp)
 	SCSI	*scgp;
 {
 	register int	f;
@@ -188,9 +254,9 @@ scsi_close(scgp)
 	if (scgp->local == NULL)
 		return (-1);
 
-	for (b=0; b < MAX_SCG; b++) {
-		for (t=0; t < MAX_TGT; t++) {
-			for (l=0; l < MAX_LUN ; l++) {
+	for (b = 0; b < MAX_SCG; b++) {
+		for (t = 0; t < MAX_TGT; t++) {
+			for (l = 0; l < MAX_LUN; l++) {
 				f = scglocal(scgp)->scgfiles[b][t][l];
 				if (f >= 0)
 					close(f);
@@ -202,7 +268,7 @@ scsi_close(scgp)
 }
 
 LOCAL BOOL
-scsi_setup(scgp, f, busno, tgt, tlun)
+scg_setup(scgp, f, busno, tgt, tlun)
 	SCSI	*scgp;
 	int	f;
 	int	busno;
@@ -215,7 +281,7 @@ scsi_setup(scgp, f, busno, tgt, tlun)
 	int	Lun;
 	BOOL	onetarget = FALSE;
 
-	if (scgp->scsibus >= 0 && scgp->target >= 0 && scgp->lun >= 0)
+	if (scg_scsibus(scgp) >= 0 && scg_target(scgp) >= 0 && scg_lun(scgp) >= 0)
 		onetarget = TRUE;
 
 	if (ioctl(f, SCIOCIDENTIFY, &saddr) < 0) {
@@ -227,8 +293,10 @@ scsi_setup(scgp, f, busno, tgt, tlun)
 	Target	= SADDR_TARGET(saddr);
 	Lun	= SADDR_LUN(saddr);
 
-	if (scgp->debug)
-		printf("Bus: %d Target: %d Lun: %d\n", Bus, Target, Lun);
+	if (scgp->debug > 0) {
+		js_fprintf((FILE *)scgp->errfile,
+			"Bus: %d Target: %d Lun: %d\n", Bus, Target, Lun);
+	}
 
 	if (Bus >= MAX_SCG || Target >= MAX_TGT || Lun >= MAX_LUN) {
 		close(f);
@@ -250,29 +318,30 @@ scsi_setup(scgp, f, busno, tgt, tlun)
 }
 
 LOCAL long
-scsi_maxdma(scgp)
+scgo_maxdma(scgp, amt)
 	SCSI	*scgp;
+	long	amt;
 {
 	long maxdma = MAX_DMA_BSD;
 
 	return (maxdma);
 }
 
-EXPORT void *
-scsi_getbuf(scgp, amt)
+LOCAL void *
+scgo_getbuf(scgp, amt)
 	SCSI	*scgp;
 	long	amt;
 {
-	if (amt <= 0 || amt > scsi_maxdma(scgp))
-		return ((void *)0);
-	if (scgp->debug)
-		printf("scsi_getbuf: %ld bytes\n", amt);
+	if (scgp->debug > 0) {
+		js_fprintf((FILE *)scgp->errfile,
+			"scgo_getbuf: %ld bytes\n", amt);
+	}
 	scgp->bufbase = valloc((size_t)(amt));
 	return (scgp->bufbase);
 }
 
-EXPORT void
-scsi_freebuf(scgp)
+LOCAL void
+scgo_freebuf(scgp)
 	SCSI	*scgp;
 {
 	if (scgp->bufbase)
@@ -280,8 +349,8 @@ scsi_freebuf(scgp)
 	scgp->bufbase = NULL;
 }
 
-EXPORT
-BOOL scsi_havebus(scgp, busno)
+LOCAL BOOL
+scgo_havebus(scgp, busno)
 	SCSI	*scgp;
 	int	busno;
 {
@@ -294,16 +363,16 @@ BOOL scsi_havebus(scgp, busno)
 	if (scgp->local == NULL)
 		return (FALSE);
 
-	for (t=0; t < MAX_TGT; t++) {
-		for (l=0; l < MAX_LUN ; l++)
+	for (t = 0; t < MAX_TGT; t++) {
+		for (l = 0; l < MAX_LUN; l++)
 			if (scglocal(scgp)->scgfiles[busno][t][l] >= 0)
 				return (TRUE);
 	}
 	return (FALSE);
 }
 
-EXPORT
-int scsi_fileno(scgp, busno, tgt, tlun)
+LOCAL int
+scgo_fileno(scgp, busno, tgt, tlun)
 	SCSI	*scgp;
 	int	busno;
 	int	tgt;
@@ -320,23 +389,22 @@ int scsi_fileno(scgp, busno, tgt, tlun)
 	return ((int)scglocal(scgp)->scgfiles[busno][tgt][tlun]);
 }
 
-EXPORT int
-scsi_initiator_id(scgp)
+LOCAL int
+scgo_initiator_id(scgp)
 	SCSI	*scgp;
 {
 	return (-1);
 }
 
-EXPORT
-int scsi_isatapi(scgp)
+LOCAL int
+scgo_isatapi(scgp)
 	SCSI	*scgp;
 
 {
 #ifdef	MAYBE_ATAPI
-	int	f = scsi_fileno(scgp, scgp->scsibus, scgp->target, scgp->lun);
 	struct scsi_addr saddr;
 
-	if (ioctl(f, SCIOCIDENTIFY, &saddr) < 0)
+	if (ioctl(scgp->fd, SCIOCIDENTIFY, &saddr) < 0)
 		return (-1);
 
 	if (!SADDR_ISSCSI(saddr))
@@ -345,28 +413,35 @@ int scsi_isatapi(scgp)
 	return (FALSE);
 }
 
-EXPORT
-int scsireset(scgp)
+LOCAL int
+scgo_reset(scgp, what)
 	SCSI	*scgp;
+	int	what;
 {
-	int	f = scsi_fileno(scgp, scgp->scsibus, scgp->target, scgp->lun);
-
-	return (ioctl(f, SCIOCRESET, 0));
+	if (what == SCG_RESET_NOP)
+		return (0);
+	if (what != SCG_RESET_BUS) {
+		errno = EINVAL;
+		return (-1);
+	}
+	/*
+	 * XXX Does this reset TGT or BUS ???
+	 */
+	return (ioctl(scgp->fd, SCIOCRESET, 0));
 }
 
 LOCAL int
-scsi_send(scgp, f, sp)
+scgo_send(scgp)
 	SCSI		*scgp;
-	int		f;
-	struct scg_cmd	*sp;
 {
+	struct scg_cmd	*sp = scgp->scmd;
 	scsireq_t	req;
 	register long	*lp1;
 	register long	*lp2;
 	int		ret = 0;
 
-/*	printf("f: %d\n", f);*/
-	if (f < 0) {
+/*	js_fprintf((FILE *)scgp->errfile, "fd: %d\n", scgp->fd);*/
+	if (scgp->fd < 0) {
 		sp->error = SCG_FATAL;
 		return (0);
 	}
@@ -387,9 +462,9 @@ scsi_send(scgp, f, sp)
 	req.databuf = sp->addr;
 	req.datalen = sp->size;
 	req.datalen_used = 0;
-	fillbytes(req.sense, sizeof(req.sense), '\0');
-	if (sp->sense_len > sizeof(req.sense))
-		req.senselen = sizeof(req.sense);
+	fillbytes(req.sense, sizeof (req.sense), '\0');
+	if (sp->sense_len > sizeof (req.sense))
+		req.senselen = sizeof (req.sense);
 	else if (sp->sense_len < 0)
 		req.senselen = 0;
 	else
@@ -399,7 +474,7 @@ scsi_send(scgp, f, sp)
 	req.retsts = 0;
 	req.error = 0;
 
-	if (ioctl(f, SCIOCCOMMAND, (void *)&req) < 0) {
+	if (ioctl(scgp->fd, SCIOCCOMMAND, (void *)&req) < 0) {
 		ret  = -1;
 		sp->ux_errno = geterrno();
 		if (sp->ux_errno != ENOTTY)
@@ -409,8 +484,8 @@ scsi_send(scgp, f, sp)
 		if (req.retsts != SCCMD_OK)
 			sp->ux_errno = EIO;
 	}
-	fillbytes(&sp->scb, sizeof(sp->scb), '\0');
-	fillbytes(&sp->u_sense.cmd_sense, sizeof(sp->u_sense.cmd_sense), '\0');
+	fillbytes(&sp->scb, sizeof (sp->scb), '\0');
+	fillbytes(&sp->u_sense.cmd_sense, sizeof (sp->u_sense.cmd_sense), '\0');
 	sp->resid = req.datalen - req.datalen_used;
 	sp->sense_count = req.senselen_used;
 	if (sp->sense_count > SCG_MAX_SENSE)
@@ -470,8 +545,8 @@ scsi_send(scgp, f, sp)
  */
 
 #undef	sense
-#define scsi_sense CAM_scsi_sense
-#define scsi_inquiry CAM_scsi_inquiry
+#define	scsi_sense CAM_scsi_sense
+#define	scsi_inquiry CAM_scsi_inquiry
 #include <sys/param.h>
 #include <cam/cam.h>
 #include <cam/cam_ccb.h>
@@ -479,23 +554,70 @@ scsi_send(scgp, f, sp)
 #include <cam/scsi/scsi_pass.h>
 #include <camlib.h>
 
-#define CAM_MAXDEVS	128
+/*
+ *	Warning: you may change this source, but if you do that
+ *	you need to change the _scg_version and _scg_auth* string below.
+ *	You may not return "schily" for an SCG_AUTHOR request anymore.
+ *	Choose your name instead of "schily" and make clear that the version
+ *	string is related to a modified source.
+ */
+LOCAL	char	_scg_trans_version[] = "scsi-bsd.c-1.42";	/* The version for this transport*/
+
+#define	CAM_MAXDEVS	128
 struct scg_local {
 	struct cam_device *cam_devices[CAM_MAXDEVS + 1];
 };
-#define scglocal(p)	((struct scg_local *)((p)->local)) 
+#define	scglocal(p)	((struct scg_local *)((p)->local))
+
+/*
+ * Return version information for the low level SCSI transport code.
+ * This has been introduced to make it easier to trace down problems
+ * in applications.
+ */
+LOCAL char *
+scgo_version(scgp, what)
+	SCSI	*scgp;
+	int	what;
+{
+	if (scgp != (SCSI *)0) {
+		switch (what) {
+
+		case SCG_VERSION:
+			return (_scg_trans_version);
+		/*
+		 * If you changed this source, you are not allowed to
+		 * return "schily" for the SCG_AUTHOR request.
+		 */
+		case SCG_AUTHOR:
+			return (_scg_auth_schily);
+		case SCG_SCCS_ID:
+			return (__sccsid);
+		}
+	}
+	return ((char *)0);
+}
+
+LOCAL int
+scgo_help(scgp, f)
+	SCSI	*scgp;
+	FILE	*f;
+{
+	__scg_help(f, "CAM", "Generic transport independent SCSI (Common Access Method)",
+		"", "bus,target,lun", "1,2,0", TRUE, FALSE);
+	return (0);
+}
 
 /*
  * Build a list of everything we can find.
  */
-EXPORT int
-scsi_open(scgp, device, busno, tgt, tlun)
+LOCAL int
+scgo_open(scgp, device)
 	SCSI	*scgp;
 	char	*device;
-	int	busno;
-	int	tgt;
-	int	tlun;
 {
+	int				busno	= scg_scsibus(scgp);
+	int				tgt	= scg_target(scgp);
+	int				tlun	= scg_lun(scgp);
 	char				name[16];
 	int				unit;
 	int				nopen = 0;
@@ -513,7 +635,7 @@ scsi_open(scgp, device, busno, tgt, tlun)
 	}
 
 	if (scgp->local == NULL) {
-		scgp->local = malloc(sizeof(struct scg_local));
+		scgp->local = malloc(sizeof (struct scg_local));
 		if (scgp->local == NULL)
 			return (0);
 
@@ -529,9 +651,9 @@ scsi_open(scgp, device, busno, tgt, tlun)
 	if (busno >= 0 && tgt >= 0 && tlun >= 0) {
 		scglocal(scgp)->cam_devices[0] = cam_open_btl(busno, tgt, tlun, O_RDWR, NULL);
 		if (scglocal(scgp)->cam_devices[0] == NULL)
-			return(-1);
+			return (-1);
 		nopen++;
-		return(nopen);
+		return (nopen);
 	}
 
 	/*
@@ -543,9 +665,9 @@ scsi_open(scgp, device, busno, tgt, tlun)
 		if (scgp->errstr)
 			js_snprintf(scgp->errstr, SCSI_ERRSTR_SIZE,
 				"Open of %s failed", XPT_DEVICE);
-		return(-1);
+		return (-1);
 	}
-	fillbytes(&ccb, sizeof(union ccb), '\0');
+	fillbytes(&ccb, sizeof (union ccb), '\0');
 
 	/*
 	 * Get a list of up to CAM_MAXDEVS passthrough devices in the
@@ -556,7 +678,7 @@ scsi_open(scgp, device, busno, tgt, tlun)
 	/*
 	 * Setup the result buffer.
 	 */
-	bufsize = sizeof(struct dev_match_result) * CAM_MAXDEVS;
+	bufsize = sizeof (struct dev_match_result) * CAM_MAXDEVS;
 	ccb.cdm.match_buf_len = bufsize;
 	ccb.cdm.matches = (struct dev_match_result *)malloc(bufsize);
 	if (ccb.cdm.matches == NULL) {
@@ -564,7 +686,7 @@ scsi_open(scgp, device, busno, tgt, tlun)
 			js_snprintf(scgp->errstr, SCSI_ERRSTR_SIZE,
 				"Couldn't malloc match buffer");
 		close(fd);
-		return(-1);
+		return (-1);
 	}
 	ccb.cdm.num_matches = 0;
 
@@ -573,20 +695,21 @@ scsi_open(scgp, device, busno, tgt, tlun)
 	 * peripherals named "pass".
 	 */
 	ccb.cdm.num_patterns = 1;
-	ccb.cdm.pattern_buf_len = sizeof(struct dev_match_pattern);
+	ccb.cdm.pattern_buf_len = sizeof (struct dev_match_pattern);
 	ccb.cdm.patterns = (struct dev_match_pattern *)malloc(
-		sizeof(struct dev_match_pattern));
+		sizeof (struct dev_match_pattern));
 	if (ccb.cdm.patterns == NULL) {
 		if (scgp->errstr)
 			js_snprintf(scgp->errstr, SCSI_ERRSTR_SIZE,
 				"Couldn't malloc pattern buffer");
 		close(fd);
 		free(ccb.cdm.matches);
-		return(-1);
+		return (-1);
 	}
 	ccb.cdm.patterns[0].type = DEV_MATCH_PERIPH;
 	match_pat = &ccb.cdm.patterns[0].pattern.periph_pattern;
-	sprintf(match_pat->periph_name, "pass");
+	js_snprintf(match_pat->periph_name, sizeof (match_pat->periph_name),
+								"pass");
 	match_pat->flags = PERIPH_MATCH_NAME;
 
 	if (ioctl(fd, CAMIOCOMMAND, &ccb) == -1) {
@@ -596,12 +719,12 @@ scsi_open(scgp, device, busno, tgt, tlun)
 		close(fd);
 		free(ccb.cdm.matches);
 		free(ccb.cdm.patterns);
-		return(-1);
+		return (-1);
 	}
 
-	if ((ccb.ccb_h.status != CAM_REQ_CMP)
-	 || ((ccb.cdm.status != CAM_DEV_MATCH_LAST)
-	    && (ccb.cdm.status != CAM_DEV_MATCH_MORE))) {
+	if ((ccb.ccb_h.status != CAM_REQ_CMP) ||
+	    ((ccb.cdm.status != CAM_DEV_MATCH_LAST) &&
+	    (ccb.cdm.status != CAM_DEV_MATCH_MORE))) {
 /*		errmsgno(EX_BAD, "Got CAM error 0x%X, CDM error %d.\n",*/
 		if (scgp->errstr)
 			js_snprintf(scgp->errstr, SCSI_ERRSTR_SIZE,
@@ -610,7 +733,7 @@ scsi_open(scgp, device, busno, tgt, tlun)
 		close(fd);
 		free(ccb.cdm.matches);
 		free(ccb.cdm.patterns);
-		return(-1);
+		return (-1);
 	}
 
 	free(ccb.cdm.patterns);
@@ -631,11 +754,12 @@ scsi_open(scgp, device, busno, tgt, tlun)
 					"Kernel error! got periph match type %d!!",
 					ccb.cdm.matches[unit].type);
 			free(ccb.cdm.matches);
-			return(-1);
+			return (-1);
 		}
 		periph_result = &ccb.cdm.matches[unit].result.periph_result;
 
-		sprintf(name, "/dev/%s%d", periph_result->periph_name,
+		js_snprintf(name, sizeof (name),
+				"/dev/%s%d", periph_result->periph_name,
 			periph_result->unit_number);
 
 		/*
@@ -672,8 +796,8 @@ scsi_open(scgp, device, busno, tgt, tlun)
 	return (nopen);
 }
 
-EXPORT int
-scsi_close(scgp)
+LOCAL int
+scgo_close(scgp)
 	SCSI	*scgp;
 {
 	register int	i;
@@ -685,13 +809,14 @@ scsi_close(scgp)
 		if (scglocal(scgp)->cam_devices[i] != (struct cam_device *)-1)
 			cam_close_device(scglocal(scgp)->cam_devices[i]);
 		scglocal(scgp)->cam_devices[i] = (struct cam_device *)-1;
-        }
+	}
 	return (0);
 }
 
-LOCAL
-long scsi_maxdma(scgp)
+LOCAL long
+scgo_maxdma(scgp, amt)
 	SCSI	*scgp;
+	long	amt;
 {
 #ifdef	DFLTPHYS
 	return (DFLTPHYS);
@@ -700,21 +825,21 @@ long scsi_maxdma(scgp)
 #endif
 }
 
-EXPORT void *
-scsi_getbuf(scgp, amt)
+LOCAL void *
+scgo_getbuf(scgp, amt)
 	SCSI	*scgp;
 	long	amt;
 {
-	if (amt <= 0 || amt > scsi_maxdma(scgp))
-		return ((void *)0);
-	if (scgp->debug)
-		printf("scsi_getbuf: %ld bytes\n", amt);
+	if (scgp->debug > 0) {
+		js_fprintf((FILE *)scgp->errfile,
+			"scgo_getbuf: %ld bytes\n", amt);
+	}
 	scgp->bufbase = valloc((size_t)(amt));
 	return (scgp->bufbase);
 }
 
-EXPORT void
-scsi_freebuf(scgp)
+LOCAL void
+scgo_freebuf(scgp)
 	SCSI	*scgp;
 {
 	if (scgp->bufbase)
@@ -722,8 +847,8 @@ scsi_freebuf(scgp)
 	scgp->bufbase = NULL;
 }
 
-EXPORT
-BOOL scsi_havebus(scgp, busno)
+LOCAL BOOL
+scgo_havebus(scgp, busno)
 	SCSI	*scgp;
 	int	busno;
 {
@@ -745,8 +870,8 @@ BOOL scsi_havebus(scgp, busno)
 	return (FALSE);
 }
 
-EXPORT
-int scsi_fileno(scgp, busno, unit, tlun)
+LOCAL int
+scgo_fileno(scgp, busno, unit, tlun)
 	SCSI	*scgp;
 	int	busno;
 	int	unit;
@@ -768,64 +893,66 @@ int scsi_fileno(scgp, busno, unit, tlun)
 	return (-1);
 }
 
-EXPORT int
-scsi_initiator_id(scgp)
+LOCAL int
+scgo_initiator_id(scgp)
 	SCSI	*scgp;
 {
 	return (-1);
 }
 
-EXPORT
-int scsi_isatapi(scgp)
+LOCAL int
+scgo_isatapi(scgp)
 	SCSI	*scgp;
 {
 	return (FALSE);
 }
 
-EXPORT
-int scsireset(scgp)
+LOCAL int
+scgo_reset(scgp, what)
 	SCSI	*scgp;
+	int	what;
 {
 	/* XXX synchronous reset command - is this wise? */
+	errno = EINVAL;
 	return (-1);
 }
 
 LOCAL int
-scsi_send(scgp, unit, sp)
-	SCSI		*scgp;
-	int		unit;
-	struct scg_cmd	*sp;
+scgo_send(scgp)
+	SCSI	*scgp;
 {
+	struct scg_cmd		*sp = scgp->scmd;
 	struct cam_device	*dev;
 	union ccb		ccb_space;
 	union ccb		*ccb = &ccb_space;
 	int			rv, result;
 	u_int32_t		ccb_flags;
 
-	if (unit < 0) {
+	if (scgp->fd < 0) {
 #if 0
-		fprintf(stderr, "attempt to reference invalid unit %d\n", unit);
+		js_fprintf((FILE *)scgp->errfile,
+			"attempt to reference invalid unit %d\n", scgp->fd);
 #endif
 		sp->error = SCG_FATAL;
 		return (0);
 	}
 
-	dev = scglocal(scgp)->cam_devices[unit];
-	fillbytes(&ccb->ccb_h, sizeof(struct ccb_hdr), '\0');
+	dev = scglocal(scgp)->cam_devices[scgp->fd];
+	fillbytes(&ccb->ccb_h, sizeof (struct ccb_hdr), '\0');
 	ccb->ccb_h.path_id = dev->path_id;
 	ccb->ccb_h.target_id = dev->target_id;
 	ccb->ccb_h.target_lun = dev->target_lun;
 
 	/* Build the CCB */
-	fillbytes(&(&ccb->ccb_h)[1], sizeof(struct ccb_scsiio), '\0');
+	fillbytes(&(&ccb->ccb_h)[1], sizeof (struct ccb_scsiio), '\0');
 	movebytes(sp->cdb.cmd_cdb, &ccb->csio.cdb_io.cdb_bytes, sp->cdb_len);
 
 	/*
 	 * Set the data direction flags.
 	 */
 	if (sp->size != 0) {
-		ccb_flags = (sp->flags & SCG_RECV_DATA) ? CAM_DIR_IN :
-							   CAM_DIR_OUT;
+		ccb_flags = (sp->flags & SCG_RECV_DATA) ?   CAM_DIR_IN :
+							    CAM_DIR_OUT;
 	} else {
 		ccb_flags = CAM_DIR_NONE;
 	}
@@ -836,21 +963,21 @@ scsi_send(scgp, unit, sp)
 	 * If you don't want to bother with sending tagged commands under CAM,
 	 * we don't need to do anything to cdrecord.  If you want to send
 	 * tagged commands to those devices that support it, we'll need to set
-	 * the tag action valid field like this in scsi_send():
- 	 *         
-	 *        ccb_flags |= CAM_DEV_QFRZDIS | CAM_TAG_ACTION_VALID;
+	 * the tag action valid field like this in scgo_send():
+	 *
+	 *	ccb_flags |= CAM_DEV_QFRZDIS | CAM_TAG_ACTION_VALID;
 	 */
 
 	cam_fill_csio(&ccb->csio,
-		      /* retries */ 1,
-		      /* cbfncp */ NULL,
-		      /* flags */ ccb_flags,
-		      /* tag_action */ MSG_SIMPLE_Q_TAG,
-		      /* data_ptr */ (u_int8_t *)sp->addr,
-		      /* dxfer_len */ sp->size,
-		      /* sense_len */ SSD_FULL_SIZE,
-		      /* cdb_len */ sp->cdb_len,
-		      /* timeout */ sp->timeout * 1000);
+			/* retries */	1,
+			/* cbfncp */	NULL,
+			/* flags */	ccb_flags,
+			/* tag_action */ MSG_SIMPLE_Q_TAG,
+			/* data_ptr */	(u_int8_t *)sp->addr,
+			/* dxfer_len */	sp->size,
+			/* sense_len */	SSD_FULL_SIZE,
+			/* cdb_len */	sp->cdb_len,
+			/* timeout */	sp->timeout * 1000);
 
 	/* Run the command */
 	errno = 0;
@@ -863,7 +990,7 @@ scsi_send(scgp, unit, sp)
 		 * error.  If we completed successfully, there's (obviously)
 		 * no error.  We declare anything else "retryable".
 		 */
-		switch(ccb->ccb_h.status & CAM_STATUS_MASK) {
+		switch (ccb->ccb_h.status & CAM_STATUS_MASK) {
 			case CAM_SEL_TIMEOUT:
 				result = SCG_FATAL;
 				break;
@@ -884,8 +1011,8 @@ scsi_send(scgp, unit, sp)
 		sp->ux_errno = EIO;
 
 	/* Pass the result back up */
-	fillbytes(&sp->scb, sizeof(sp->scb), '\0');
-	fillbytes(&sp->u_sense.cmd_sense, sizeof(sp->u_sense.cmd_sense), '\0');
+	fillbytes(&sp->scb, sizeof (sp->scb), '\0');
+	fillbytes(&sp->u_sense.cmd_sense, sizeof (sp->u_sense.cmd_sense), '\0');
 	sp->resid = ccb->csio.resid;
 	sp->sense_count = SSD_FULL_SIZE - ccb->csio.sense_resid;
 
@@ -905,6 +1032,6 @@ scsi_send(scgp, unit, sp)
 
 #undef scsi_sense
 #undef scsi_inquiry
-#define sense u_sense.Sense
+#define	sense u_sense.Sense
 
 #endif /* BSD_CAM */
